package myutil;

import java.util.Deque;
import java.util.LinkedList;
import java.util.Random;

import aiinterface.CommandCenter;
import enumerate.Action;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;

/**
 * MCTS?利用?るNode
 *
 * @author Taichi Miyazaki
 */
public class Node {

	/** UCT?実行時間 */
	// public static final int UCT_TIME = 165 * 100000;
	public long UCT_TIME;
	/** UCB1?定数C?値 */
	public static final double UCB_C = 3;

	/** 探索?る木?深? */
	public static final int UCT_TREE_DEPTH = 2;

	/** ノードを生??る閾値 */
	public static final int UCT_CREATE_NODE_THRESHOULD = 10;

	/** シミュレーションを行?時間 */
	public static final int SIMULATION_TIME = 60;

	/** 乱数を利用?る????使? */
	Random rnd;

	/** 親ノード */
	Node parent;

	/** ?ノード */
	Node[] children;

	/** ノード?深? */
	int depth;

	/** ノード?探索?れ?回数 */
	int games;

	/** UCB1値 */
	double ucb;

	/** 評価値 */
	double score;

	/** ?択???る自分?全Action */
	LinkedList<Action> myActions;

	/** ?択???る相手?全Action */
	LinkedList<Action> oppActions;

	/** シミュレーション?る????利用?る */
	Simulator simulator;

	/** 探索時??ん?自分?Action */
	LinkedList<Action> selectedMyActions;

	/** シミュレーション?る??自分?HP */
	int myOriginalHp;

	/** シミュレーション?る??相手?HP */
	int oppOriginalHp;

	FrameData frameData;
	boolean playerNumber;
	CommandCenter commandCenter;
	GameData gameData;

	boolean isCreateNode;

	Deque<Action> mAction;
	Deque<Action> oppAction;

	public Node(long ucttime, FrameData frameData, Node parent, LinkedList<Action> myActions,
			LinkedList<Action> oppActions, GameData gameData, boolean playerNumber, CommandCenter commandCenter,
			LinkedList<Action> selectedMyActions) {
		this(ucttime, frameData, parent, myActions, oppActions, gameData, playerNumber, commandCenter);

		this.selectedMyActions = selectedMyActions;
	}

	public Node(long ucttime, FrameData frameData, Node parent, LinkedList<Action> myActions,
			LinkedList<Action> oppActions, GameData gameData, boolean playerNumber, CommandCenter commandCenter) {
		this.UCT_TIME = ucttime;
		this.frameData = frameData;
		this.parent = parent;
		this.myActions = myActions;
		this.oppActions = oppActions;
		this.gameData = gameData;
		this.simulator = new Simulator(gameData);
		this.playerNumber = playerNumber;
		this.commandCenter = commandCenter;

		this.selectedMyActions = new LinkedList<Action>();

		this.rnd = new Random();
		this.mAction = new LinkedList<Action>();
		this.oppAction = new LinkedList<Action>();

		CharacterData myCharacter = frameData.getCharacter(playerNumber);
		CharacterData oppCharacter = frameData.getCharacter(!playerNumber);
		myOriginalHp = myCharacter.getHp();
		oppOriginalHp = oppCharacter.getHp();

		if (this.parent != null) {
			this.depth = this.parent.depth + 1;
		} else {
			this.depth = 0;
		}
	}

	/**
	 * MCTSを行?
	 *
	 * @return 最終的?ノード?探索回数?多?Action
	 */
	public Action mcts() {
		// 時間??り?UCTを繰り返?
		long start = System.nanoTime();
		for (; System.nanoTime() - start <= UCT_TIME;) {
			uct();
		}

		return getBestVisitAction();
	}

	/**
	 * プレイアウト(シミュレーション)を行?
	 *
	 * @return プレイアウト?果?評価値
	 */
	public double playout() {

		mAction.clear();
		oppAction.clear();

		for (int i = 0; i < selectedMyActions.size(); i++) {
			mAction.add(selectedMyActions.get(i));
		}

		for (int i = 0; i < 5 - selectedMyActions.size(); i++) {
			mAction.add(myActions.get(rnd.nextInt(myActions.size())));
		}

		for (int i = 0; i < 5; i++) {
			Action toac;
			toac = oppActions.get(rnd.nextInt(oppActions.size()));
			oppAction.add(oppActions.get(rnd.nextInt(oppActions.size())));
		}

		FrameData nFrameData = simulator.simulate(frameData, playerNumber, mAction, oppAction, SIMULATION_TIME); // シミュレーションを実行

		return getScore(nFrameData);
	}

	/**
	 * UCTを行? <br>
	 *
	 * @return 評価値
	 */
	public double uct() {

		Node selectedNode = null;
		double bestUcb;

		bestUcb = -99999;

		for (Node child : this.children) {
			if (child.games == 0) {
				child.ucb = 9999 + rnd.nextInt(50);
			} else {
				child.ucb = getUcb(child.score / child.games, games, child.games);
			}

			if (bestUcb < child.ucb) {
				selectedNode = child;
				bestUcb = child.ucb;
			}

		}

		double score = 0;
		if (selectedNode.games == 0) {
			score = selectedNode.playout();
		} else {
			if (selectedNode.children == null) {
				if (selectedNode.depth < UCT_TREE_DEPTH) {
					if (UCT_CREATE_NODE_THRESHOULD <= selectedNode.games) {
						selectedNode.createNode();
						selectedNode.isCreateNode = true;
						score = selectedNode.uct();
					} else {
						score = selectedNode.playout();
					}
				} else {
					score = selectedNode.playout();
				}
			} else {
				if (selectedNode.depth < UCT_TREE_DEPTH) {
					score = selectedNode.uct();
				} else {
					selectedNode.playout();
				}
			}

		}

		selectedNode.games++;
		selectedNode.score += score;

		if (depth == 0) {
			games++;
		}

		return score;
	}

	/**
	 * ノードを生??る
	 */
	public void createNode() {

		this.children = new Node[myActions.size()];

		for (int i = 0; i < children.length; i++) {

			LinkedList<Action> my = new LinkedList<Action>();
			for (Action act : selectedMyActions) {
				my.add(act);
			}

			my.add(myActions.get(i));

			children[i] = new Node(this.UCT_TIME, frameData, this, myActions, oppActions, gameData, playerNumber,
					commandCenter, my);
		}
	}

	public void createNode(LinkedList<Action> myacs) {

		this.children = new Node[myacs.size()];

		for (int i = 0; i < children.length; i++) {

			LinkedList<Action> my = new LinkedList<Action>();
			for (Action act : selectedMyActions) {
				my.add(act);
			}

			my.add(myacs.get(i));

			children[i] = new Node(this.UCT_TIME, frameData, this, myActions, oppActions, gameData, playerNumber,
					commandCenter, my);
		}
	}

	/**
	 * 最多訪?回数?ノード?Actionを返?
	 *
	 * @return 最多訪?回数?ノード?Action
	 */
	public Action getBestVisitAction() {

		int selected = -1;
		double bestGames = -9999;

		for (int i = 0; i < children.length; i++) {

			if (bestGames < children[i].games) {
				bestGames = children[i].games;
				selected = i;
			}
		}

		return children[selected].selectedMyActions.getFirst();
		// return this.myActions.get(selected);
	}

	/**
	 * 評価値を返?
	 *
	 * @param fd
	 *            フレームデータ(?れ?hp???情報?入???る)
	 * @return 評価値
	 */
	public int getScore(FrameData fd) {
		if (playerNumber) {
			return (fd.getCharacter(true).getHp() - myOriginalHp) - (fd.getCharacter(false).getHp() - oppOriginalHp);
		} else {
			return (fd.getCharacter(false).getHp() - myOriginalHp) - (fd.getCharacter(true).getHp() - oppOriginalHp);
		}

	}

	/**
	 * 評価値?全プレイアウト試行回数????Action?プレイアウト試行回数?らUCB1値を返?
	 *
	 * @param score
	 *            評価値
	 * @param n
	 *            全プレイアウト試行回数
	 * @param ni
	 *            ???Action?プレイアウト試行回数
	 * @return UCB1値
	 */
	public double getUcb(double score, int n, int ni) {
		return score + UCB_C * Math.sqrt((2 * Math.log(n)) / ni);
	}

	public void printNode(Node node) {
		System.out.println("全試行回数:" + node.games);
		for (int i = 0; i < node.children.length; i++) {
			System.out.println(i + ",回数:" + node.children[i].games + ",深?:" + node.children[i].depth + ",score:"
					+ node.children[i].score / node.children[i].games + ",ucb:" + node.children[i].ucb);
		}
		System.out.println("");
		for (int i = 0; i < node.children.length; i++) {
			if (node.children[i].isCreateNode) {
				printNode(node.children[i]);
			}
		}
	}
}
